/* $Workfile: T2pE1ABCDBit.c $       $Revision: 45 $      $Modtime: 5/26/10 6:11p $  */
/*******************************************************************************

  Odin TeleSystems Inc.

  OTX Telecom Adapter Driver

  @doc PublicDemo

  @module

    T2pE1ABCDBit.c

    A simple example console application demonstrating the use of OTX Driver
    to access T1 Rob-Bit signalling bits with following boards:

      Thor-2-PCI and Thor-8-PCI

    or OTX PCI Plus series boards,

      Thor-2-PCI-Plus, Thor-4-PCI-Plus and Thor-8-PCI-Plus 

    or OTX PCI Express series boards,

      Thor-2-PCI-Express

    or Thor-2-PCMCIA series boards,

      Thor-2-PCMCIA, 
      Thor-2-PCMCIA-Plus, 
      Thor-2-PCMCIA-Ex, 
      Thor-2-PCMCIA-Pro,
      Thor-2-PCMCIA-CST

    or Sleipnir-1-PCI-Plus

    or Alvis-2-CSI
    or Alvis-4-CSI or Alvis-4M-CSI
    or Alvis-8-CSI or Alvis-8M-CSI

    The sample application provides a simple console interface which utilizes
    simple keystrokes as commands.

    It configures 2 T1 Lines and expects them to be looped with an external
    loop cable. The T1 lines are used for accessing Bit Robbed Channel
    Associated Signaling information.

    2 Handsets are connected to the codecs and a through connection is
    demonstrated through the Loop Cable.

    Press 'h' or '?' for help. 'q' to quit.

*******************************************************************************/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#if !defined(_WIN32_WCE)
#include <assert.h>
#include <time.h>
#endif  // #if !defined(_WIN32_WCE)

#include "OsaStd.h"
#include "OsaCon.h"
#include "OsaThrd.h"
#include "OtxErr.h"
#include "OtxBrd.h"
#include "OtxDef.h"
#include "OtxType.h"
#include "OtxDrv.h"
#include "OtxT1E1.h"
#include "OtxData.h"

#define DEMO_DEFAULT_LOG_FILE "T2pE1ABCDBit.log"
#define DEMO_DEFAULT_BOARD_NO 0

typedef struct DemoEnvT {
  char            *pszInputStr;
  OTX_BOOL        bExitOnError;
  char            cwd[OTX_MAX_PATH];
  char            logFileName[OTX_MAX_PATH];
  OTX_INT32       nBoardType;
  int             coLi;
  int             boardNo;
  OTX_BOOL        bChannelizedAbcdBitAccess; // If OTX_TRUE then report ABCD data per channel in the OTX_T1E1_EC_BITROB_SIGNALING event
} DemoEnvT;

/*******************************************************************************
 Function Prototypes
*******************************************************************************/
static void DemoCreateEventQueue(void);
static void DemoOpenPhysicalDevices(void);
static void DemoConfigureE1T1Devices(void);
static void DemoCreateLogicalDevices(void);
static void DemoConnectDevices(void);
static void DemoInitializeLogicalDevices(void);
static void DemoEnableDevices(void);
static void DemoInitT1E1ConfigData(OUT OtxE1ConfigOptionsS *pE1CfgData, OUT OtxT1ConfigOptionsS *pT1CfgData);
static void DemoPrintEvent(IN OtxEventDataS *pEventData);

static void DemoParseCmdLineArgs(int argc, char *argv[], DemoEnvT *envP);
static void DemoHelpPanel(void);
static void DemoLogf(char *formatStr, ...);
static void DemoCheckRc(IN OTX_RESULT otxRc);
static void DemoExitQuestion(void);

static OTX_INT32 DemoProcessUserCommand(char commandChar);


/*******************************************************************************
 Global Data
*******************************************************************************/
OTX_EVENT   g_hNotificationEvent = OTX_INVALID_EVENT_VALUE;
OTX_HANDLE  g_hEventQueue  = 0;
FILE       *g_pDemoLogFile = NULL;
DemoEnvT    g_demoEnv;
OtxE1ABCDBitDataS  g_sqSendABCDBitData[8];

struct PhysDevs {
  OTX_HANDLE hThorController0;
  OTX_HANDLE hQuadE1T1[2];        // Used for Thor-4-PCI-Plus and Thor-8-PCI(-Plus)
  OTX_HANDLE hLi[8];
} g_hPhysDevs;


struct Pipes {
  OTX_HANDLE hLiIn[8];
  OTX_HANDLE hLiOut[8];
} g_hPipes;


struct LogicDevs {
  OTX_HANDLE hT1E1Cas[8];
} g_hLogicDevs;


struct T1E1CfgData {
  OtxE1ConfigOptionsS e1CfgData;
  OtxT1ConfigOptionsS t1CfgData;
} g_T1E1CfgData;


/*******************************************************************************
*******************************************************************************/


/*******************************************************************************

  @func

  Create the event queue to be used by the driver to inform the application
  about events, such as line status changes, or detected DMTF digits.

*******************************************************************************/
void DemoCreateEventQueue(void)
{
  OTX_RESULT result;

  // Create a notification Event
  result = OtxDrvCreateEvent(&g_hNotificationEvent);
  if (OTX_RESULT_CODE(result) == OTX_E_NOTIMPL) {
    g_hNotificationEvent = (OTX_EVENT) NULL;    // D16V86 apps cannot use event objects
    result = OTX_S_OK;
  }

  DemoCheckRc(result);

  // Create Event Queue for driver notifications
  result = OtxDrvCreateEventQueue(128, g_hNotificationEvent, &g_hEventQueue); DemoCheckRc(result);

  // Enable Event Queue
  result = OtxDrvEnable(g_hEventQueue, OTX_TASK_SYNC,0); DemoCheckRc(result);
}



/*******************************************************************************

  @func
    Open the Physical Devices on the board

  @comm
    Calls function <f OtxDrvCreatePhysicalDevice> returning a handle to a
    Device Object representing a physical device, which can be the
    entire board or an integrated circuit (chip) on the board. The number of
    physical devices is set per board and cannot change dynamically.

*******************************************************************************/
void DemoOpenPhysicalDevices(void)
{
  OTX_RESULT result;
  OTX_INT32  i;
  OTX_INT32  nQuadE1T1; // Only used for Thor-4-PCI-Plus, Thor-8-PCI(-Plus)


  // Connect to the Physical Board Controller
  switch (g_demoEnv.nBoardType) {
  case OTX_DEVICE_THOR_PCI:
    DemoLogf("Opening Thor-2-PCI board\n");
    g_demoEnv.coLi = 2;
    break;
  case OTX_DEVICE_THOR_2_PCI_PLUS:
    DemoLogf("Opening Thor-2-PCI_PLUS board\n");
    g_demoEnv.coLi = 2;
    break;
  case OTX_DEVICE_THOR_2_PCI_EXPRESS:
    DemoLogf("Opening Thor-2-PCI-Express board\n");
    g_demoEnv.coLi = 2;
    break;
  case OTX_DEVICE_THOR_4_PCI_PLUS:
    DemoLogf("Opening Thor-4-PCI-Plus board\n");
    g_demoEnv.coLi = 4;
    break;
  case OTX_DEVICE_THOR_8_PCI:
    DemoLogf("Opening Thor-8-PCI board\n");
    g_demoEnv.coLi = 8;
    break;
  case OTX_DEVICE_THOR_PCMCIA:
    DemoLogf("Opening Thor-2-PCMCIA variant board\n");
    g_demoEnv.coLi = 2;
    break;
  case OTX_DEVICE_THOR_8_PCI_PLUS:
    DemoLogf("Opening Thor-8-PCI-Plus board\n");
    g_demoEnv.coLi = 8;
    break;
  case OTX_DEVICE_SLEIPNIR_1_PCI:
    DemoLogf("Opening Sleipnir-1-PCI board\n");
    g_demoEnv.coLi = 1;
    break;
  case OTX_DEVICE_THOR_2_EXPRESSCARD:
    DemoLogf("Opening Thor-2-ExpressCard board\n");
    g_demoEnv.coLi = 2;
    break;
  case OTX_DEVICE_DMP_ALVIS_2_CSI:
    DemoLogf("Opening Alvis-2-CSI board\n");
    g_demoEnv.coLi = 2;
    break;
  case OTX_DEVICE_DMP_ALVIS_4_CSI:
    DemoLogf("Opening Alvis-4-CSI board\n");
    g_demoEnv.coLi = 4;
    break;
  case OTX_DEVICE_DMP_ALVIS_8_CSI:
    DemoLogf("Opening Alvis-8-CSI board\n");
    g_demoEnv.coLi = 8;
    break;
  case OTX_DEVICE_DMP_ALVIS_4M_CSI:
    DemoLogf("Opening Alvis-4M-CSI board\n");
    g_demoEnv.coLi = 4;
    break;
  case OTX_DEVICE_DMP_ALVIS_8M_CSI:
    DemoLogf("Opening Alvis-8M-CSI board\n");
    g_demoEnv.coLi = 8;
    break;
  default:
    DemoLogf("Board Type (%d) not supported by this demo program\n", g_demoEnv.nBoardType);
    exit(1);
    break;
  }
  result = OtxDrvOpenPhysicalDevice(0, (OtxPhysicalDeviceTypeE) g_demoEnv.nBoardType, g_demoEnv.boardNo, g_demoEnv.boardNo, g_hEventQueue, &g_hPhysDevs.hThorController0); DemoCheckRc(result);

  // Get handles to T1/E1 line interface devices #0 and #1
  switch (g_demoEnv.nBoardType) {
  case OTX_DEVICE_THOR_PCI:
  case OTX_DEVICE_THOR_PCMCIA:
  case OTX_DEVICE_SLEIPNIR_1_PCI:
  case OTX_DEVICE_THOR_2_PCI_PLUS:
  case OTX_DEVICE_THOR_2_PCI_EXPRESS:
    // On these boards the E1T1 devices are hosted by the board itself. Get handles to T1/E1 line interface devices #0 and #1
    for (i = 0; i < g_demoEnv.coLi; i++) {
      DemoLogf("Open OTX_DEVICE_LI_T1E1 #%d\n", i);
      result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hThorController0, OTX_DEVICE_LI_T1E1, i, i, g_hEventQueue, &g_hPhysDevs.hLi[i]); DemoCheckRc(result);
    }
    break;
  case OTX_DEVICE_THOR_4_PCI_PLUS:
  case OTX_DEVICE_DMP_ALVIS_4_CSI:
  case OTX_DEVICE_DMP_ALVIS_4M_CSI:
  case OTX_DEVICE_DMP_ALVIS_8_CSI:
  case OTX_DEVICE_DMP_ALVIS_8M_CSI:
    {
      // On the Thor-4-PCI-Plus board the E1T1 devices are hosted by a quad E1T1 device. 
      for (nQuadE1T1 = 0; nQuadE1T1 < 1; nQuadE1T1++) {
        DemoLogf("Open OTX_DEVICE_QUAD_T1E1 #%d\n", nQuadE1T1);
        result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hThorController0, OTX_DEVICE_QUAD_T1E1, nQuadE1T1, nQuadE1T1, g_hEventQueue, &g_hPhysDevs.hQuadE1T1[nQuadE1T1]); DemoCheckRc(result);
      }
      
      // Get handles to T1/E1 line interface devices 
      for (i = 0; i < g_demoEnv.coLi; i++) {
        nQuadE1T1 = i/4;
        DemoLogf("Open OTX_DEVICE_LI_T1E1 #%d (hosted by OTX_DEVICE_QUAD_T1E1 #%d)\n", i, nQuadE1T1);
        result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hQuadE1T1[nQuadE1T1], OTX_DEVICE_LI_T1E1, i%4, i, g_hEventQueue, &g_hPhysDevs.hLi[i]); DemoCheckRc(result);
      }
    }
    break;
  case OTX_DEVICE_THOR_8_PCI:
  case OTX_DEVICE_THOR_8_PCI_PLUS:
    // On the Thor-8-PCI(-Plus) board the E1T1 devices are hosted by a quad E1T1 device. Open those two devices.
    for (nQuadE1T1 = 0; nQuadE1T1 < 2; nQuadE1T1++) {
      DemoLogf("Open OTX_DEVICE_QUAD_T1E1 #%d\n", nQuadE1T1);
      result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hThorController0, OTX_DEVICE_QUAD_T1E1, nQuadE1T1, nQuadE1T1, g_hEventQueue, &g_hPhysDevs.hQuadE1T1[nQuadE1T1]); DemoCheckRc(result);
    }

    // Get handles to T1/E1 line interface devices 
    for (i = 0; i < g_demoEnv.coLi; i++) {
      nQuadE1T1 = i/4;
      DemoLogf("Open OTX_DEVICE_LI_T1E1 #%d (hosted by OTX_DEVICE_QUAD_T1E1 #%d)\n", i, nQuadE1T1);
      result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hQuadE1T1[nQuadE1T1], OTX_DEVICE_LI_T1E1, i%4, i, g_hEventQueue, &g_hPhysDevs.hLi[i]); DemoCheckRc(result);
    }
    break;
  case OTX_DEVICE_DMP_ALVIS_2_CSI:
  case OTX_DEVICE_THOR_2_EXPRESSCARD:
    // On the Thor-2-ExpressCard board the E1T1 devices are hosted by a dual E1T1 device. 
    DemoLogf("Open OTX_DEVICE_DUAL_T1E1\n");
    result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hThorController0, OTX_DEVICE_DUAL_T1E1, 0, 0, g_hEventQueue, &g_hPhysDevs.hQuadE1T1[0]); DemoCheckRc(result);

    // Get handles to T1/E1 line interface devices 
    for (i = 0; i < g_demoEnv.coLi; i++) {
      DemoLogf("Open OTX_DEVICE_LI_T1E1 #%d (hosted by OTX_DEVICE_QUAD_T1E1)\n", i);
      result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hQuadE1T1[0], OTX_DEVICE_LI_T1E1, i, i, g_hEventQueue, &g_hPhysDevs.hLi[i]); DemoCheckRc(result);
    }
    break;
  default:
    DemoLogf("Board Type (%d) not supported by this demo program\n", g_demoEnv.nBoardType);
    exit(1);
    break;
  }

}



/*******************************************************************************

  @func
    Configure E1/T1 Devices

  @comm The clock source of the board(s) must set by a call 
  to <f OtxBrdSetClocks> prior to calling this function.

*******************************************************************************/
void DemoConfigureE1T1Devices(void)
{
  OTX_RESULT result;
  OTX_INT32  i;

  // Set default configuration data
  DemoInitT1E1ConfigData( &(g_T1E1CfgData.e1CfgData), &(g_T1E1CfgData.t1CfgData));

  // Configure Transceivers
  for (i = 0; i < g_demoEnv.coLi; i++) {

    // Set termination for the T1/E1 Front End
    result = OtxT1E1SetLineTermination(g_hPhysDevs.hLi[i], OTX_T1E1_TERM_100_OHM); DemoCheckRc(result);

    result = OtxE1SetConfiguration(g_hPhysDevs.hLi[i], &(g_T1E1CfgData.e1CfgData));  DemoCheckRc(result);
    result = OtxT1E1SetLiMode(g_hPhysDevs.hLi[i], OTX_T1E1_MODE_E1); DemoCheckRc(result);

    if (g_demoEnv.bChannelizedAbcdBitAccess) {
      result = OtxDrvSetAttributeValueBool(g_hPhysDevs.hLi[i], OTX_ATTR_T1E1_CHANNELIZED_ABCD_BIT_ACCESS, OTX_TRUE); DemoCheckRc(result);
    }
  }

}



/*******************************************************************************
  @func
    Create the needed Logical Devices
  @comm
    The logical devices are differenciated from the phyiscal devices by the fact
    the number of logical devices is not set. More of them can be requested
    by the user and allocated by the system if needed.
*******************************************************************************/
void DemoCreateLogicalDevices(
  void
)
{
  OTX_RESULT result;
  OTX_INT32  i;

  for (i = 0; i < g_demoEnv.coLi; i++) {
    // Create CAS Sender and Receiver logical devices hosted in the physical Line Interface devices
    result = OtxDrvCreateLogicalDevice( g_hPhysDevs.hLi[i], OTX_LDEVICE_E1_ABCD_BIT_ACCESS, g_hEventQueue, i, &g_hLogicDevs.hT1E1Cas[i]); DemoCheckRc(result);
  }

}


/*******************************************************************************

  @func
    Create the needed Pipe Devices and connect the physical and logical devices
    to pipes.

  @comm
    A Pipe Device is a logical entity representing a serial bit-stream (a subset of a
    Highway) with a data rate between 8 kbit/s and 2.048Mbit/s. A pipe is a characterized
    bu its capacity (in 8 kbit/s increments). The user can create a pipe with
    the driver function <f OtxDrvCreatePipe>. A pipe is connected to physical or
    logical devices and mapped to specific time-slots with the driver function
    <f OtxDrvConnectPipe>.

*******************************************************************************/
void DemoConnectDevices(
  void
)
{
  OTX_INT32 i;
  OTX_RESULT result;
  OTX_UINT8 sqTimeSlotMask[32];
  OTX_INT32 ts;

  for (i = 0; i < g_demoEnv.coLi; i++) {
    // Create an incoming and outgoing Pipe representing the external T1/E1
    result = OtxDrvCreatePipe(64, &g_hPipes.hLiIn[i]); DemoCheckRc(result);
    result = OtxDrvCreatePipe(64, &g_hPipes.hLiOut[i]); DemoCheckRc(result);

    // Connect external T1/E1 pipes. Only connect one end of each pipe. External Interface is LI Highway Number 1
    ts = 24;
    memset(sqTimeSlotMask, 0, 32 * sizeof(OTX_UINT8));
    sqTimeSlotMask[ts] = 0xFF;
    result = OtxDrvConnectPipe(g_hPipes.hLiIn[i], OTX_PIPE_OUTPUT, g_hPhysDevs.hLi[i], 1, sqTimeSlotMask, 32); DemoCheckRc(result);
    result = OtxDrvConnectPipe(g_hPipes.hLiOut[i], OTX_PIPE_INPUT, g_hPhysDevs.hLi[i], 1, sqTimeSlotMask, 32); DemoCheckRc(result);

  }
}


/*******************************************************************************
  @func
    Initialize the logical device. This function must be called after
    the devices are created and connected.
*******************************************************************************/
void DemoInitializeLogicalDevices(
  void
)
{
  return;
}


/*******************************************************************************

  @func
    Enable Configured Devices

*******************************************************************************/
void DemoEnableDevices(void)
{
  OTX_RESULT result;
  OTX_INT32  i;
  OTX_INT32  nQuadE1T1;

  // Enable Physical Devices
  result = OtxDrvEnable(g_hPhysDevs.hThorController0,  OTX_TASK_SYNC, 5000L); DemoCheckRc(result);

  // If board type is Thor-4-PCI-Plus or Thor-2-ExpressCard or Alvis-4-CSI, then enable Quad E1T1 device
  if ( (g_demoEnv.nBoardType == OTX_DEVICE_THOR_4_PCI_PLUS) || (g_demoEnv.nBoardType == OTX_DEVICE_THOR_2_EXPRESSCARD) || (g_demoEnv.nBoardType == OTX_DEVICE_DMP_ALVIS_2_CSI)  || (g_demoEnv.nBoardType == OTX_DEVICE_DMP_ALVIS_4_CSI) || (g_demoEnv.nBoardType == OTX_DEVICE_DMP_ALVIS_4M_CSI)) {
      
    // Enable Quad E1T1 devices
    for (nQuadE1T1 = 0; nQuadE1T1 < 1; nQuadE1T1++) {
      DemoLogf("Enable OTX_DEVICE_QUAD_T1E1 #%d device\n", nQuadE1T1);
      result = OtxDrvEnable(g_hPhysDevs.hQuadE1T1[nQuadE1T1], OTX_TASK_SYNC, 5000L); DemoCheckRc(result);
    }
  }

  // If board type is Thor-8-PCI(-Plus) then enable Quad E1T1 device
  if (g_demoEnv.nBoardType == OTX_DEVICE_THOR_8_PCI || g_demoEnv.nBoardType == OTX_DEVICE_THOR_8_PCI_PLUS || g_demoEnv.nBoardType == OTX_DEVICE_DMP_ALVIS_8_CSI || g_demoEnv.nBoardType == OTX_DEVICE_DMP_ALVIS_8M_CSI) {
    // Enable Quad E1T1 devices
    for (nQuadE1T1 = 0; nQuadE1T1 < 2; nQuadE1T1++) {
      DemoLogf("Enable OTX_DEVICE_QUAD_T1E1 #%d device\n", nQuadE1T1);
      result = OtxDrvEnable(g_hPhysDevs.hQuadE1T1[nQuadE1T1], OTX_TASK_SYNC, 5000L); DemoCheckRc(result);
    }
  }

  for (i = 0; i < g_demoEnv.coLi; i++) {
    result = OtxDrvEnable(g_hPhysDevs.hLi[i],            OTX_TASK_SYNC, 5000L); DemoCheckRc(result);
  }

  // Enable Logical T1/E1 Cas Devices
  for (i = 0; i < g_demoEnv.coLi; i++) {
    result = OtxDrvEnable(g_hLogicDevs.hT1E1Cas[i],   OTX_TASK_SYNC, 5000L); DemoCheckRc(result);
  }

}


/*******************************************************************************

  @func
    Initialize the T1/E1 Configuration Data Structures passed to the T1/E1
    Line Interface physical devices

*******************************************************************************/
void DemoInitT1E1ConfigData(
  OUT OtxE1ConfigOptionsS *pE1CfgData,
  OUT OtxT1ConfigOptionsS *pT1CfgData
)
{
  // Initialize T1 Configuration Data
  pT1CfgData->eSignalingMode             = OTX_T1_SM_CAS_BR;
  pT1CfgData->eTxLineCode                = OTX_T1_LC_AMI;
  pT1CfgData->eRxLineCode                = OTX_T1_LC_AMI;
  pT1CfgData->eFrameFormat               = OTX_T1_FF_ESF;
  pT1CfgData->bCRC6Enable                = OTX_TRUE;
  pT1CfgData->eTxRemoteAlarmFormat       = OTX_T1_RA_YELLOW_A;
  pT1CfgData->eRxRemoteAlarmFormat       = OTX_T1_RA_YELLOW_A;
  pT1CfgData->bAutoResync                = OTX_TRUE;
  pT1CfgData->bAutoCrcBitInversion       = OTX_FALSE;
  pT1CfgData->eLfaSensitivity            = OTX_T1_LFA_2_OUT_OF_4;
  pT1CfgData->bAutoRemoteAlarm           = OTX_FALSE;
  pT1CfgData->nLosSensitivity            = 10;  // 10 = LOS detection after 176 consecutive zeros which fulfills G.775, Telcordia (Bellcore) and AT&T
  pT1CfgData->nLosRecovery               = 21;  // 21 = LOS recovery after 22 ones in the LosRecovery interval which fulfills G.775, Telcordia (Bellcore) and AT&T
  pT1CfgData->nLineLength                = 10;
  pT1CfgData->bHighTxPower               = OTX_FALSE;
  pT1CfgData->bReceiveEqualizer          = OTX_TRUE;
  pT1CfgData->bfClearChannels            = 0xC00000; // 0xC00000 means that TS23 and TS24 are marked as clear channels. The data in these timeslots will not be overwritten by bitrobbing data.

  // Initialize E1 Configuration Data
  pE1CfgData->eTxLineCode                = OTX_E1_LC_HDB3;
  pE1CfgData->eRxLineCode                = OTX_E1_LC_HDB3;
  pE1CfgData->eTxFrameFormat             = OTX_E1_FF_CRC4_MULTIFRAME;
  pE1CfgData->eRxFrameFormat             = OTX_E1_FF_CRC4_MULTIFRAME;
  pE1CfgData->eAisDetection              = OTX_E1_AIS_ETS300233;
  pE1CfgData->bfSiBits                   = 0x3;
  pE1CfgData->bfSaBits                   = 0x1F;
  pE1CfgData->bExtendedHDB3errorDetect   = OTX_FALSE;
  pE1CfgData->bAutoRegainMultiframe      = OTX_FALSE;
  pE1CfgData->bAutoResync                = OTX_TRUE;
  pE1CfgData->bAutoCrcBitInversion       = OTX_FALSE;
  pE1CfgData->bAutoRemoteAlarm           = OTX_FALSE;
  pE1CfgData->nLosSensitivity            = 10;  // 10 = LOS detection after 176 consecutive zeros which fulfills G.775
  pE1CfgData->nLosRecovery               = 21;  // 21 = LOS recovery after 22 ones in the LosRecovery interval which fulfills G.775
  pE1CfgData->nLineLength                = 10;
  pE1CfgData->bHighTxPower               = OTX_FALSE;
  pE1CfgData->bReceiveEqualizer          = OTX_FALSE;
}





/*******************************************************************************
  @func
    Idle Function that check for Events from the OTX Driver
*******************************************************************************/
void *DemoIdleFunction(
  void *lpParameter   // thread data
)
{
  OtxEventDataS eventData;
  OTX_RESULT result;

  while (1) {
    OtxDrvWaitForSingleEvent(g_hNotificationEvent, OTX_TIME_INFINITY);
    //DemoLogf("Event signalled\n");
    do {
      result = OtxDrvGetEventData(g_hEventQueue, &eventData); DemoCheckRc(result);

      if (OTX_RESULT_CODE(result) == OTX_S_OK) {
        // We have a driver event
        // Printf the event data.
        // In a real application branch based on the data
        DemoPrintEvent(&eventData);

      }
    } while (OTX_RESULT_CODE(result) == OTX_S_OK);
  }

  return 0;

}



/*******************************************************************************
  @func
    Prints a Driver Event.
*******************************************************************************/
void DemoPrintEvent(
  OtxEventDataS *pEventData
)
{
  OTX_RESULT otxRc;
  char msgStr[256];

  DemoLogf("%.8X%.8X ", pEventData->m_nTimeStamp.sUint32.nHigh, pEventData->m_nTimeStamp.sUint32.nLow);

  switch (pEventData->m_eDeviceType) {
  case OTX_DEVICE_TSS:
    // Event from the Time-Space Switch
    otxRc = OtxDrvEventCode2String(pEventData->m_hDevice, pEventData->m_nCode, 256, &msgStr[0]);
    if (OTX_SUCCESS(otxRc)) {
      DemoLogf("TSS Event: %s \n", msgStr);
    } else {
      DemoLogf("TSS Invalid Event: %ld \n", pEventData->m_nCode);
    }
    break;

  case OTX_DEVICE_LI_T1E1:
    // Event from Li0 or Li1
    otxRc = OtxDrvEventCode2String(pEventData->m_hDevice, pEventData->m_nCode, 256, &msgStr[0]);
    if (OTX_FAILURE(otxRc)) {
      DemoLogf("E1/T1[%d] No string for event: %ld \n", pEventData->m_nSourceId, pEventData->m_nCode);
    }
    if (pEventData->m_nParam) {
      DemoLogf("E1/T1[%d] Event: %s (0x%x)\n", pEventData->m_nSourceId, msgStr, pEventData->m_nParam);
    } else {
      DemoLogf("E1/T1[%d] Event: %s\n", pEventData->m_nSourceId, msgStr);
    }

    break;

  case OTX_LDEVICE_E1_ABCD_BIT_ACCESS:
    otxRc = OtxDrvEventCode2String(pEventData->m_hDevice, pEventData->m_nCode, 256, &msgStr[0]);
    if (OTX_FAILURE(otxRc)) {
      DemoLogf("ABCD Device[%d] No String for event: %ld \n", pEventData->m_nSourceId, pEventData->m_nCode);
    }

    if (pEventData->m_nCode == OTX_T1E1_EC_BITROB_SIGNALING) {
      if (g_demoEnv.bChannelizedAbcdBitAccess) {
        DemoLogf("ABCD Device[%d] Event: %s (Channel:%.2d ABCD:0x%X)\n", pEventData->m_nSourceId, msgStr, pEventData->m_nParam >> 8, pEventData->m_nParam & 0x0F);
      } else {
        DemoLogf("ABCD Device[%d] %c Bits changed (0x%0.8x)\n", pEventData->m_nSourceId, 'A' + (pEventData->m_nParam >> 30), pEventData->m_nParam & 0x3fffffff);
      }
    } else {

      if (pEventData->m_nParam) {
        DemoLogf("ABCD Device[%d] Event: %s (0x%x)\n", pEventData->m_nSourceId, msgStr, pEventData->m_nParam);
      } else {
        DemoLogf("ABCD Device[%d] Event: %s\n", pEventData->m_nSourceId, msgStr);
      }
    }
    break;


  default:
    // Unhandled event
    DemoLogf("Event from an unknown object 0x%X\n", pEventData->m_hDevice);
    break;
  }


}


/*******************************************************************************
  @func
    Extract a specific channel ABCD bit value from the OtxE1ABCDBitDataS struct.
*******************************************************************************/
char *DemoGetCasChannelValue(
  IN OTX_INT32 nCasCh,
  IN OtxE1ABCDBitDataS  *pABCDBitData,
  OUT char *pszValue  // @parm ABCD bits (in that order) as a string
)
{
  pszValue[0] = '\0';
  
  if (nCasCh < 1 || nCasCh > 30) {
    DemoLogf("Error: CAS Channel out of range (%d)\n", nCasCh);
    return pszValue;
  }

  sprintf(pszValue + strlen(pszValue), "%d",  (pABCDBitData->chA >> (nCasCh - 1)) & 0x01 ? 1 : 0);
  sprintf(pszValue + strlen(pszValue), "%d",  (pABCDBitData->chB >> (nCasCh - 1)) & 0x01 ? 1 : 0);
  sprintf(pszValue + strlen(pszValue), "%d",  (pABCDBitData->chC >> (nCasCh - 1)) & 0x01 ? 1 : 0);
  sprintf(pszValue + strlen(pszValue), "%d",  (pABCDBitData->chD >> (nCasCh - 1)) & 0x01 ? 1 : 0);

  return pszValue;
}


/*******************************************************************************
  @func
    Sets a specific CAS-channel ABCD bit value in the OtxE1ABCDBitDataS struct.
*******************************************************************************/
char *DemoSetCasChannelValue(
  IN OTX_INT32 nCasCh,
  IN OtxE1ABCDBitDataS *pABCDBitData,
  IN char *pszValue  // @parm ABCD bits (in that order) as a string
)
{
  OTX_INT32 i;
  OTX_UINT32 *pChX = &(pABCDBitData->chA);
  
  if (nCasCh < 1 || nCasCh > 30) {
    DemoLogf("Error: CAS Channel out of range (%d)\n", nCasCh);
    return pszValue;
  }

  for (i = 0; i < 4; i++) {
    switch (i) {
    case 0:
      pChX = &(pABCDBitData->chA);
      break;
    case 1:
      pChX = &(pABCDBitData->chB);
      break;
    case 2:
      pChX = &(pABCDBitData->chC);
      break;
    case 3:
      pChX = &(pABCDBitData->chD);
      break;
    }

    switch (pszValue[i]) {
    case '1':
      *pChX |=   0x01 << (nCasCh - 1);
      break;
    case '0':
      *pChX &= ~(0x01 << (nCasCh - 1));
      break;
    default:
      DemoLogf("Error: Invalid bit value (%c). Expected '0' or '1'\n", pszValue[i]);
      return pszValue;
    }
  }

  return pszValue;
}
  

/*******************************************************************************
  @func
    Prints the ABCD Bit data for all 30 CAS channels
*******************************************************************************/
void DemoPrintAllCasChannelsData(
  OTX_INT32 nCasDev
)
{
  OTX_RESULT result;

  DemoLogf("Reading received bitrob data on LI#%d ... \n", nCasDev);

  if (g_demoEnv.bChannelizedAbcdBitAccess) {
    OTX_INT32 i;
    OTX_UINT8 nABCDBits;

    for (i = 1; i <= 30; i++) {
      result = OtxE1GetChannelizedABCDBit(g_hLogicDevs.hT1E1Cas[nCasDev], i, &nABCDBits); DemoCheckRc(result);
      DemoLogf("   Channel %.2d: 0x%X", i, nABCDBits);
      if (!(i % 2)) {
        DemoLogf("\n");
      }
    }

  } else {
    OTX_INT32 i;
    OtxE1ABCDBitDataS  rxABCDBitData;
    char szABCDbuf[4+1];  // +1 for the '\0'
    
    result = OtxE1GetABCDBit(g_hLogicDevs.hT1E1Cas[nCasDev], &rxABCDBitData); DemoCheckRc(result);
    if (OTX_RESULT_CODE(result) != OTX_S_OK) {
      return;
    }
    DemoLogf("A = %.6X, B = %.6X, C = %.6X, D = %.6X\n", rxABCDBitData.chA, rxABCDBitData.chB, rxABCDBitData.chC, rxABCDBitData.chD);
    DemoLogf("\n");
    DemoLogf("            ABCD              ABCD\n");
    DemoLogf("            ====              ====\n");
    for (i = 1; i <= 30; i++) {
      DemoLogf("Channel %.2d: ", i);
      DemoLogf("%s", DemoGetCasChannelValue(i, &rxABCDBitData, szABCDbuf));
      DemoLogf("  ");
      if (!(i % 2)) {
        DemoLogf("\n");
      }
    }
    DemoLogf("\n");
  } 
}




/*******************************************************************************
  @func
    Parse the program's command line arguments and calls upon
    requested actions.

    There are some optional switches that can be used following
    the driver name on the command line. Switches are are
    separated by spaces, and can appear in arbitrary order.
    Switches consist of a hyphen and a letter:
*******************************************************************************/
static void DemoParseCmdLineArgs(int argc, char *argv[], DemoEnvT *envP) {
  short  i;

  // Set reasonable defaults before reading from command line
  envP->pszInputStr               = NULL;
  envP->bExitOnError              = OTX_FALSE;
  envP->boardNo                   = DEMO_DEFAULT_BOARD_NO;
  envP->nBoardType                = OTX_DEVICE_UNKNOWN;
  envP->coLi                      = 2;
  envP->bChannelizedAbcdBitAccess = OTX_TRUE;

  strcpy(envP->logFileName, envP->cwd);
  strcat(envP->logFileName, DEMO_DEFAULT_LOG_FILE);
  for (i = 1; i < argc; i++) {        // Scan command line arguments
    if (*(argv[i]+1) == 'Z') {   // Command input string -- case sensitive
      envP->pszInputStr = argv[i]+2;
    } else {    
      OsaStdStrUpr(argv[i]);
      if (*argv[i] == '-' || *argv[i] == '/') {  // Is argument a legal switch?
        if (*(argv[i]+1) == 'L') {   // Log file
          strcpy(envP->logFileName, argv[i] + 2);
        } else if (*(argv[i]+1) == 'Q') {   // Exit on error
          envP->bExitOnError = OTX_TRUE;
        } else if (*(argv[i]+1) == 'B') {   // Board number
          envP->boardNo = strtoul(argv[i]+2, NULL, 0);
        } else if (*(argv[i]+1) == 'U') {   // Un-Channelized ABCD bit access
          envP->bChannelizedAbcdBitAccess = OTX_FALSE;
        } else if (*(argv[i]+1) == 'T') {   // Board Type
          if ( *(argv[i]+2) == '2') {
            envP->nBoardType = OTX_DEVICE_THOR_PCI;
          } else if ( *(argv[i]+2) == 'T') {
            envP->nBoardType  = OTX_DEVICE_THOR_2_PCI_PLUS;
          } else if ( *(argv[i]+2) == 'E') {
            envP->nBoardType  = OTX_DEVICE_THOR_2_PCI_EXPRESS;
          } else if ( *(argv[i]+2) == '4') {
            envP->nBoardType  = OTX_DEVICE_THOR_4_PCI_PLUS;
          } else if ( *(argv[i]+2) == '8') {
            envP->nBoardType  = OTX_DEVICE_THOR_8_PCI;
          } else if ( *(argv[i]+2) == 'M') {
            envP->nBoardType  = OTX_DEVICE_THOR_PCMCIA;
          } else if ( *(argv[i]+2) == 'P') {
            envP->nBoardType  = OTX_DEVICE_THOR_8_PCI_PLUS;
          } else if ( *(argv[i]+2) == 'S') {
            envP->nBoardType  = OTX_DEVICE_SLEIPNIR_1_PCI;
          } else if ( *(argv[i]+2) == 'C') {
            envP->nBoardType  = OTX_DEVICE_THOR_2_EXPRESSCARD;
          } else if ( *(argv[i]+2) == 'D') {
            envP->nBoardType  = OTX_DEVICE_DMP_ALVIS_4_CSI;
          } else if ( *(argv[i]+2) == 'K') {
            envP->nBoardType  = OTX_DEVICE_DMP_ALVIS_4M_CSI;
          } else if (*(argv[i] + 2) == 'X') {
            envP->nBoardType  = OTX_DEVICE_DMP_ALVIS_8_CSI;
          } else if (*(argv[i] + 2) == 'Q') {
            envP->nBoardType  = OTX_DEVICE_DMP_ALVIS_8M_CSI;
          } else if (*(argv[i] + 2) == 'B') {
            envP->nBoardType  = OTX_DEVICE_DMP_ALVIS_2_CSI;
          } else {
            DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
            DemoHelpPanel();
            exit(1);
          }
        
        } else {
          DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
          DemoHelpPanel();
          OsaConsoleExit();
				  exit(1);
        }
      } else {
        DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
        DemoHelpPanel();
        OsaConsoleExit();
			  exit(1);
      }
    }
  }

}


/*******************************************************************************
  @func
    Displays a help for the command line.
*******************************************************************************/
static void DemoHelpPanel()
{
  DemoLogf("\nCommand line switches:\n");
  DemoLogf(" /B<boardNo>    : Board number (Default is %d)\n", DEMO_DEFAULT_BOARD_NO);
  DemoLogf(" /L<fileName>   : Log file name\n");
  DemoLogf(" /T[2|T|E|4|8|M|P|S|C|D|K|X|Q|B]: Board Type used for the demo, where\n");
  DemoLogf("                    2 = Use Thor-2-PCI Board (default)\n");
  DemoLogf("                    T = Use Thor-2-PCI-Plus Board\n");
  DemoLogf("                    E = Use Thor-2-PCI-Express Board\n");
  DemoLogf("                    4 = Use Thor-4-PCI-Plus Board\n");
  DemoLogf("                    8 = Use Thor-8-PCI Board\n");
  DemoLogf("                    M = Use Thor-2-PCMCIA Board\n");
  DemoLogf("                    P = Use Thor-8-PCI-Plus Board\n");
  DemoLogf("                    S = Use Sleipnir-1-PCI Board\n");
  DemoLogf("                    C = Use Thor-2-ExpressCard Board\n");
  DemoLogf("                    D = Use Alvis-4-CSI Board\n");
  DemoLogf("                    K = Use Alvis-4M-CSI Board\n");
  DemoLogf("                    X = Use Alvis-8-CSI Board\n");
  DemoLogf("                    Q = Use Alvis-8M-CSI Board\n");
  DemoLogf("                    B = Use Alvis-2-CSI Board\n");
  DemoLogf(" /U             : Use un-channelized ABCD bit access\n");
  DemoLogf(" /Z<str>        : Command string (overrides keyboard input)\n");
  DemoLogf(" /Q             : Exit without prompting on errors\n");

  DemoLogf("\nPress Any Key to Exit...\n"); OsaConsoleGetCh();

}


/*******************************************************************************
  @func
    Displays a message on the screen and prints it to the logfile.
*******************************************************************************/
static void DemoLogf(char *formatStr, ...)
{
  va_list argList;
  va_start(argList, formatStr);

  OsaConsolePrintF(formatStr, argList);
  if (g_pDemoLogFile) vfprintf(g_pDemoLogFile,   formatStr, argList);

  va_end(argList);

}


static void DemoCheckRc(IN OTX_RESULT result)
/*******************************************************************************
  @func
    Check the result from API call. If the result is a failure code then prompt
    the user if he wants to exit.
*******************************************************************************/
{
  if ( OTX_FAILURE(result) ) {
    OTX_RESULT otxRc;
    char msgStr[256];
    
    // Convert the result code to a more descriptive string
    otxRc = OtxDrvResultCode2String(OTX_DRIVER_DEVICE_HANDLE, result, 256, &msgStr[0]);
    if (OTX_FAILURE(otxRc)) {
      strcpy(msgStr, "Cannot translate this result code.");
    }
    DemoLogf("ERROR(%#lX [%d.%d]): %s \n", OTX_RESULT_CODE(result), OTX_RESULT_SOURCE_ID(result), OTX_RESULT_LINE_NUM(result), msgStr);
    
    DemoExitQuestion();
  }
}


/*******************************************************************************
  @func
    Prompts whether to exit the program. Usually called after an error 
    has been reported.

*******************************************************************************/
void DemoExitQuestion()
{
  short c;

  if (g_demoEnv.bExitOnError) {
    c = 'y';
  } else {
    DemoLogf("Do you want to exit (Y/N) ");
    c = OsaConsoleGetCh();
  }
  if (c == 'y' || c == 'Y') {
    // Clean up
    OtxDrvReleaseLib();
    OsaConsoleExit();
    exit(1);
  }
}


/*******************************************************************************
  @func
    Processes User Commands
*******************************************************************************/
OTX_INT32 DemoProcessUserCommand(char commandChar)
{
  static OTX_INT32 nCurrentCasChannel = 1;
  static OTX_INT32 nCurrentLi = 0;
  OTX_RESULT result;

  switch (commandChar) {

  case '?':
  case 'h':
  case 'H':
    DemoLogf("Cmd  Synopsis\n");
    DemoLogf("===  ========\n");
    DemoLogf(" ?   Display Help text\n");
    DemoLogf(" h   Ditto\n");
    DemoLogf(" q   Quit\n");
    DemoLogf(" 0-7 Select active Li, this line is then used for subsequent commands\n");
    DemoLogf(" n   Increment current CAS channel (currently channel %d selected\n", nCurrentCasChannel);
    DemoLogf(" N   Decrement current CAS channel (currently channel %d selected\n", nCurrentCasChannel);
    DemoLogf(" z   Set ABCD-bits to ABCD = 0101 on all CAS channels on current Li\n");
    DemoLogf(" s   Set ABCD-bits for current CAS channel on current Li (prompt for value)\n");
    if (g_demoEnv.coLi > 1) {
      DemoLogf(" r   Read the value of the received ABCD bit data on the other Li\n");
    }
    DemoLogf(" R   Read the value of the received ABCD bit data on the current Li\n");
    DemoLogf(" l   Query Line Status of current Li\n");
    break;

    case 'z':
      {
        DemoLogf("Setting transmitted ABCD bit data (for all channels) to 0101 (0x55) on LI#%d \n", nCurrentLi);
        if (g_demoEnv.bChannelizedAbcdBitAccess) {
          OTX_INT32 i;
          for (i = 1; i <= 30; i++) {
            result = OtxE1SetChannelizedABCDBit(g_hLogicDevs.hT1E1Cas[nCurrentLi], i, 0x05); DemoCheckRc(result);
          }
        } else {

          g_sqSendABCDBitData[nCurrentLi].chA = 0x00000000;
          g_sqSendABCDBitData[nCurrentLi].chB = 0xffffffff;
          g_sqSendABCDBitData[nCurrentLi].chC = 0x00000000;
          g_sqSendABCDBitData[nCurrentLi].chD = 0xffffffff;
          result = OtxE1SetABCDBit(g_hLogicDevs.hT1E1Cas[nCurrentLi], &g_sqSendABCDBitData[nCurrentLi]); DemoCheckRc(result);
        }
        DemoLogf("Done\n");
      }
      break;


    case 's':
      {
        char szABCDbuf[4+1];  // +1 for the '\0'
        char szBuf[12];  // +1 for the '\0'

        if (g_demoEnv.bChannelizedAbcdBitAccess) {
          OTX_UINT8 nABCDBits;

          DemoLogf("Enter ABCD-bit data (0x0 - 0xF) for CAS channel %d on Li%d: ", nCurrentCasChannel, nCurrentLi);
          OsaConsoleGetStr(szBuf);
          nABCDBits = (OTX_UINT8)strtoul(szBuf, NULL, 0);
          result = OtxE1SetChannelizedABCDBit(g_hLogicDevs.hT1E1Cas[nCurrentLi], nCurrentCasChannel, nABCDBits); DemoCheckRc(result);

        } else {
          DemoGetCasChannelValue(nCurrentCasChannel, &g_sqSendABCDBitData[nCurrentLi], szABCDbuf);
          
          DemoLogf("Enter ABCD-bit data for CAS channel %d on Li%d (ABCD=%s): ", nCurrentCasChannel, nCurrentLi, szABCDbuf);
          OsaConsoleGetStr(szBuf);
          if (*szBuf) {
            if (strlen(szBuf) != 4) {
              DemoLogf("Invalid entry (%s). Please enter ABCD-bit data as a 4 character string.\n", szBuf);
              DemoLogf("E.g. \"0000\", or \"0101\" \n");
              break;
            } else {
              DemoSetCasChannelValue(nCurrentCasChannel, &g_sqSendABCDBitData[nCurrentLi], szBuf);
              result = OtxE1SetABCDBit(g_hLogicDevs.hT1E1Cas[nCurrentLi], &g_sqSendABCDBitData[nCurrentLi]); DemoCheckRc(result);
            }
          } else {
            DemoLogf("ABCD-bit data remains unchanged for this channel\n");
          }
          DemoLogf("Done.\n");
        }
      }
      break;


    case 'r':
      if (g_demoEnv.coLi > 1) {
        OTX_INT32 nLi = nCurrentLi + 1;
        
        if ((nCurrentLi % 2) == 0) {
          // Even Li
          nLi = nCurrentLi + 1;
        } else {
          // Odd Li
          nLi = nCurrentLi - 1;
        }
        DemoPrintAllCasChannelsData(nLi);
      }
      break;

    case 'R':
      {
        DemoPrintAllCasChannelsData(nCurrentLi);
      }
      break;


  case 'l':
    {
      OtxT1E1Layer1StatusE eLayer1Status;
      DemoLogf("Line status of Li #%d is ", nCurrentLi);

      result = OtxT1E1GetLayer1Status(g_hPhysDevs.hLi[nCurrentLi], &eLayer1Status); //DemoCheckRc(result);

      DemoLogf("%s\n", eLayer1Status == OTX_T1E1_LAYER1_OK ? "OK" : "DOWN");
    }
    break;

  case 'n':
    {
      // Select current nCurrentCasChannel
      nCurrentCasChannel++;
      if (nCurrentCasChannel > 30) {
        // Wrap 
        nCurrentCasChannel = 1;
      }
      
      DemoLogf("CAS Channel #%d Selected \n", nCurrentCasChannel);
      
    }
    break;
    
    
  case 'N':
    {
      // Select current nCurrentCasChannel
      nCurrentCasChannel--;
      if (nCurrentCasChannel < 1) {
        // Wrap
        nCurrentCasChannel = 30;
      }
      
      DemoLogf("CAS Channel #%d Selected \n", nCurrentCasChannel);
    }
    break;
    

  case 'i':
    {
      
      DemoLogf("Idlecode FF on CurrentLi\n", nCurrentLi);
      // set the idle character for the line
      result = OtxDrvSetAttributeValueUint32(g_hPhysDevs.hLi[nCurrentLi], OTX_ATTR_T1E1_TX_IDLE_CHARACTER, 0xFF); DemoCheckRc(result);
      result = OtxDrvSetAttributeValueUint32(g_hPhysDevs.hLi[nCurrentLi], OTX_ATTR_T1E1_USE_TX_IDLE_CHARACTER, OTX_TRUE); DemoCheckRc(result);
    
    }
    break;

  case 'j':
    {
      
      DemoLogf("Idlecode 1F on CurrentLi\n", nCurrentLi);
      // set the idle character for the line
      result = OtxDrvSetAttributeValueUint32(g_hPhysDevs.hLi[nCurrentLi], OTX_ATTR_T1E1_TX_IDLE_CHARACTER, 0x1F); DemoCheckRc(result);
      result = OtxDrvSetAttributeValueUint32(g_hPhysDevs.hLi[nCurrentLi], OTX_ATTR_T1E1_USE_TX_IDLE_CHARACTER, OTX_TRUE); DemoCheckRc(result);
    
    }
    break;

  case '0':
  case '1':
  case '2':
  case '3':
  case '4':
  case '5':
  case '6':
  case '7':
    {
      // Select current Li
      OTX_INT32 nLi = commandChar - '0';

      if (nLi > g_demoEnv.coLi) {
        DemoLogf("Error: LI#%d is not supported by this board\n", nLi);
      } else {
        nCurrentLi = nLi;
        DemoLogf("LI#%d Selected ....\n", nCurrentLi);
      }
    }
    break;

  case '~':
    {
      char strBuf[12];
      OTX_INT32 nSleepTime;

      DemoLogf("Enter milliseconds to sleep :");
      OsaConsoleGetStr(strBuf);

      if (strBuf[0]) {
        nSleepTime = strtol(strBuf, NULL, 0);
        DemoLogf("Sleeping %d milliseconds\n", nSleepTime);
        OsaThreadSleep(nSleepTime);
      }
    }
    break;

  case 'q':
  case 'Q':
    return 1;
    break;

  default:
    DemoLogf("Unknown command. Press '?' for help, 'q' to quit\n");
    break;
  }

  return 0;
}


/*******************************************************************************

  @func

  Enumerates and finds the first board device

*******************************************************************************/
static OTX_RESULT FindBoardDevice(OTX_INT32 *peBrdType) 
{
  OTX_UINT32   eDevType;
  OTX_ITERATOR itrChildTypes;
  OTX_ITERATOR itrChildrenOfSameType;
  static char szBuf[256];
  OTX_RESULT otxRc;

  *peBrdType = OTX_DEVICE_UNKNOWN; // Defensive
  
  // Initialize the opaque iterator for the hosted types
  OtxDrvIteratorBegin(&itrChildTypes);
  while (OTX_SUCCESS(OtxDrvEnumDeviceType(0, &itrChildTypes, (OTX_INT32 *)&eDevType))) {   // Note that we must use a zero handle to get the children of the root level
    OTX_HANDLE hChildDevice;
    // Initialize the opaque iterator for the devices within the same type
    OtxDrvIteratorBegin(&itrChildrenOfSameType);
    while (OTX_SUCCESS(OtxDrvEnumDevice(0, eDevType, &itrChildrenOfSameType, &hChildDevice))) {
      if (hChildDevice != OTX_INVALID_HANDLE_VALUE) {

        // Check device type
        switch (eDevType) {
          // These devices appear at the top (driver level) but they are not board devices, so ignore them here
        case OTX_DEVICE_OTPD_INTERFACE:
        case OTX_DEVICE_CT_BUS:
          break;
        default:
          otxRc = OtxDrvGetAttributeValueSz(hChildDevice, OTX_ATTR_DEVICE_STRING, szBuf, 256);
          if (!OTX_SUCCESS(otxRc)) {
            DemoLogf("Could not get the OTX_ATTR_DEVICE_STRING attribute for device 0x%X\n", hChildDevice);
            return otxRc;
          } 
          DemoLogf("Found %s board (devType=%d)\n", szBuf, eDevType);
          *peBrdType = eDevType;
          return OTX_S_OK;
        }

      }
    }
  }

  return OTX_S_FALSE; // We did not find an OTX board
}


/*******************************************************************************
*******************************************************************************/
int main(int argc, char *argv[])
{
  OTX_INT32 exitFlag = 0;
  OTX_RESULT result;  // OTX Standard Return Code
  char commandChar;
  OTX_INT32 i;

	OsaConsoleInit();

  DemoLogf("T2pE1ABCDBit, %s, Odin TeleSystems Inc.\n", __DATE__);
  DemoLogf("Example program for accessing E1 ABCD-bit data\n");
  DemoParseCmdLineArgs(argc, argv, &g_demoEnv);

  if (g_demoEnv.pszInputStr) {
    OsaConsoleInput(g_demoEnv.pszInputStr);
  }

  if ((g_pDemoLogFile = fopen(g_demoEnv.logFileName, "wt")) == NULL) {
    printf("Couldn't open logfile %s\n", g_demoEnv.logFileName);
    OsaConsoleExit();
		exit(-1);
  }

  DemoLogf("Log file is  %s\n", g_demoEnv.logFileName);
  DemoLogf("Board number %d\n", g_demoEnv.boardNo);
  DemoLogf("ABCD Bit access is %s\n", g_demoEnv.bChannelizedAbcdBitAccess ? "channelized" : "non-channelized");

  for (i = 0; i < g_demoEnv.coLi; i++) {
    g_sqSendABCDBitData[i].chA = 0x00000000;
    g_sqSendABCDBitData[i].chB = 0xffffffff;
    g_sqSendABCDBitData[i].chC = 0x00000000;
    g_sqSendABCDBitData[i].chD = 0xffffffff;
  }


  // Query Library revision
  {
    char szVersionInfo[255];
    OTX_BOOL bIsStaticLinkedLibrary;
    char szLibFileName[64];
    result = OtxDrvGetAttributeValueSz(OTX_LIB_HANDLE, OTX_ATTR_LIB_REV_STR, szVersionInfo, 255); DemoCheckRc(result);
    result = OtxDrvGetAttributeValueBool(OTX_LIB_HANDLE, OTX_ATTR_LIB_TYPE_STATIC, &bIsStaticLinkedLibrary); DemoCheckRc(result);
    result = OtxDrvGetAttributeValueSz(OTX_LIB_HANDLE, OTX_ATTR_LIB_FILE_NAME, szLibFileName, 64); DemoCheckRc(result);
    
    DemoLogf("Library version %s (filename %s, %s linked)\n", szVersionInfo, szLibFileName, bIsStaticLinkedLibrary ? "static":"dynamic");
  }


  // Connect to the Driver
  DemoLogf("Connecting lib...\n");
  result = OtxDrvConnectLib(); 
  switch (OTX_RESULT_CODE(result)) {
  case OTX_RES_HANDLE:
    DemoLogf("Error: Could not connect to the OTX library.\nPossible cause: OTX Driver not loaded\n");
    OsaConsoleExit();
    OtxDrvReleaseLib();
    exit(1);
    break;
  case OTX_RES_NOT_SUPPORTED:
    DemoLogf("Error: Mismatch between OTX Driver revision and OTX Library revision.\nRevisions must match.\n");
    OsaConsoleExit();
    OtxDrvReleaseLib();
    exit(1);
    break;
  default:
    DemoCheckRc(result);
    break;
  }

  if (g_demoEnv.nBoardType == OTX_DEVICE_UNKNOWN) {
    // Board type was not supplied by a command line parameter, so enumerate and find a board
    result = FindBoardDevice(&g_demoEnv.nBoardType);
    if (result == OTX_S_FALSE) {
      DemoLogf("No OTX board was found\n");
      DemoExitQuestion();
    } else if (result != OTX_S_OK) {
      DemoExitQuestion();
    }
  }

  // Create Event Queues for Driver --> Application Communications
  DemoLogf("Creating event queue...\n");
  DemoCreateEventQueue();
  DemoLogf("Created event queue\n");

  // Connect (get handles) to the physical devices on the Thor-2-PCI board
  DemoLogf("Opening physical devices...\n");
  DemoOpenPhysicalDevices();
  DemoLogf("Opened physical devices\n");

  // Set clock source for the board (important to set the clock source before configuring the E1/T1 devices)
  result = OtxBrdSetClocks(g_hPhysDevs.hThorController0, OTX_CLOCK_MODE_LOCAL, OTX_CLOCK_SOURCE_INTERNAL); DemoCheckRc(result);

  // Configure E1/T1 devices
  DemoLogf("Configuring E1/T1 devices...\n");
  DemoConfigureE1T1Devices();
  DemoLogf("Configured E1/T1 devices\n");

  // Create and get handles to the Logical devices needed by this application
  DemoLogf("Creating logical devices...\n");
  DemoCreateLogicalDevices();
  DemoLogf("Created logical devices\n");

  // Create and get handles to the Pipe devices needed by this application
  DemoConnectDevices();

  // Initialized the Connected Logical Devices
  DemoInitializeLogicalDevices();

  // Enable all devices. Their device state should now be OTX_DEVICE_WORKING
  DemoEnableDevices();


  // Read Serial Number from the board
  {
    OtxAttribTypeE  eAttribType;    // Type of the attribute, if attribute exists
    OTX_UINT32      cbAttribValue;  // Size in bytes of attribute value (Get value with <f OtxDrvGetAttributeValue>)
    OTX_BOOL        bReadOnly;      // Is attribute writable or read-only
    char *pszSerialNumber;

    result = OtxDrvGetAttributeType(g_hPhysDevs.hThorController0, OTX_ATTR_BRD_SERIAL_NUMBER, &eAttribType, &cbAttribValue, &bReadOnly); DemoCheckRc(result);
    assert(eAttribType == OTX_ATTRIB_SZ);
    assert(bReadOnly == OTX_TRUE);
    pszSerialNumber = (char *) malloc(cbAttribValue * sizeof(char));
    result = OtxDrvGetAttributeValueSz(g_hPhysDevs.hThorController0, OTX_ATTR_BRD_SERIAL_NUMBER, pszSerialNumber, cbAttribValue); DemoCheckRc(result);
    DemoLogf("Board Serial Number: %s\n", pszSerialNumber);
    free(pszSerialNumber);
  }

  // Main Loop. Loop here and wait for Events from the driver or commands from the user ....
	OsaThreadStart(DemoIdleFunction);

  OsaThreadSleep(300);
  DemoLogf("\nPress 'h' for help\n");

  while (!exitFlag) {
    commandChar = OsaConsoleGetCh();
    exitFlag = DemoProcessUserCommand(commandChar);
  }

  // Clean up
  OtxDrvReleaseLib();
	OsaConsoleExit();

  return 0;
}

#if defined(_WIN32_WCE)
int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPWSTR lpCmdLine, int nShowCmd) 
{
  // default command line parameters
  WCHAR wsExeName[OTX_MAX_PATH * sizeof(WCHAR)];
  static char exename[OTX_MAX_PATH];
  static char *cmdParams[] = { exename, "/TM", NULL }; 
  char *pc;

  GetModuleFileName(NULL, wsExeName, sizeof(wsExeName));
  WideCharToMultiByte(CP_ACP, 0, &wsExeName[0], sizeof(wsExeName), exename, sizeof(exename), NULL, NULL); 
  OsaStrCpy(g_demoEnv.cwd, cmdParams[0]);
  pc = OsaStrRChr(g_demoEnv.cwd, '\\');
  if (pc != NULL) {
    *++pc = '\0';
  }

  return OsaWinCeConsoleMain(hInstance, hPrevInstance, lpCmdLine, nShowCmd,
              cmdParams, sizeof(cmdParams) / sizeof(cmdParams[0]), main);
}
#endif  // #if defined(_WIN32_WCE)
